{
  Roland Potter - Woodstock Computer Solutions
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.
}
{ - These are structures used for building datasets from flat file sources.
}
unit WcsFlatFileType;

interface
uses
  Windows, SysUtils, Classes, WcsDataObjects, WCSStreamer;
const
  cFFListVers = 10;
{ Streaming sub-versions
  Version 8
    Consolidate all changes, move to bi-directional stream, leave code to read old streams.
  Version 9
    Add fFamily, fRecDelim, fFldDelim array of max supported types per version.
  Version 10
    Additional delimited bits
}
type
  TFieldDefinitionList = Class;

// This is the actual storage type in the source data file.
// These will be added to as mmore file types are catered for.
// It will be qualified with a byte-length and
// implicit/explicit DP and sign if numeric.
  TWCSFlatFileSrcType =
     (wffs_Chars     // General alpha
     ,wffs_NumPk02   // Packed Numeric (base 2 storage)
     ,wffs_NumPk10   // Packed Numeric (Nibble per digit base 10 storage)
     ,wffs_NumUnPk   // UnPacked Numeric (byte per digit base 10 storage)
     ,wffs_Single    // Ansi/ieee single precision float
     ,wffs_Double    // Ansi/ieee double precision float
     ,wffs_Binary    // All other data
     );
  TWCSFlatFileBestScanType =
     (wffbs_Any
     ,wffbs_Chars
     ,wffbs_Date
     ,wffbs_Float
     ,wffbs_Integer
     );

// This is how the data will be interprteted for the initial display pass,
// and for data extraction to achieve a minimal set of displayable/storable types.
// Only the Chars and BCD types have any length info.
// There is then a further mapping from these types to the final database storage type.
// Default is defined by the mapping routine....Mask will be used if supplied.
  TWCSFlatFileUsgType =
    (wffu_Chars     // any number of characters.
    ,wffu_LongInt   // 1-4 byte Integer.
    ,wffu_LargeInt  // 5-8 byte Integer.
    ,wffu_Float     // Floating point.
    ,wffu_BCD       // Large numerics that will not fit in Integer or Float.
    ,wffu_DateTime  // All dates and times - NEED MASKS!
    ,wffu_Binary    // Anything else!
    );

  TDatePartType = (dptYear, dptMon, dptDay, dptHour, dptMin, dptSec);
  TDateMaskRec = record
    Posns: array[TDatePartType] of Byte;
    YearLen     : Byte;
    Valid, Seps : Boolean;
  end;

// This is applied to both SIGN and DECIMAL info
// Either Exp type can be used if decimal
  TWCSFlatFileNumFlag = (nfNone, nfImp, nfExpLead, nfExpTrail);

// Global translation flags
  TWCSFlatFileDTD = record
    HiToLo    : Boolean;  // Integer Byte order - false => Lo to Hi - overrides default
    Ebsidic   : Boolean;  // Byte to char mapping - false => ansi
    WordStore : Boolean;  // Data must be on 2, 4, 8, 16, 32 etc byte boundaries
  end;

  TFieldDefinition = class(TWcsDataObject)
  private
    fDList: TFieldDefinitionList;
    fDIndex: integer;
    fDName: String;          // original name
    fDSrcPic: String;        // original specification (for ref only)
    fDSrcSize: integer;      // The total length in Bytes in the source file
    fDSrcType: TWCSFlatFileSrcType; // The basic source type

    fDSrcPrec: integer; // TOTAL Length in DECIMAL digits (not including dp/sign)
    fDSrcDP  : integer; // Number of DP accuracy stored (<= Prec)
    fDSrcSignStorage : TWCSFlatFileNumFlag;
    fDSrcDPStorage   : TWCSFlatFileNumFlag;

    // Translation Mask from Source to Dest e.g.
    // 9(6) time might be "hhmmss"
    // S9(9) COMP date might be be "yyyymmdd"
    fDSrcDstMask : string;

    // The VERY basic type for the Flat File Dataset
    // For more complex types - map onto the required type in the final destination
    // OR Use persistent fields and change their type.
    fDUsgType: TWCSFlatFileUsgType;
    // Destination does not have an absolute size (dictated by internal storage)

    // Map to final output field
    fDDstID  : Longint;
    fDSrcFilter: String;

    function GetdName: String;
    procedure SetDName(const Value: String);
    procedure SetDSrcSize(const Value: integer);
    procedure SetDSrcType(const Value: TWCSFlatFileSrcType);
    procedure SetDUsgType(const Value: TWCSFlatFileUsgType);
    procedure SetDDstID(const Value: Longint);
    procedure SetDSrcDP(const Value: integer);
    procedure SetDSrcDPStorage(const Value: TWCSFlatFileNumFlag);
    procedure SetDSrcDstMask(const Value: string);
    procedure SetDSrcPrec(const Value: integer);
    procedure SetDSrcSignStorage(const Value: TWCSFlatFileNumFlag);
    procedure SetDSrcPic(const Value: String);
    procedure SetDSrcFilter(const Value: String);
  protected
    procedure StreamValuesOLD(const Stream: TStream); override;
    function GetDataObjName: String; override;
  public
    constructor Create(const AList: TFieldDefinitionList); reintroduce;
    property DList: TFieldDefinitionList read fDList;
    procedure Init; override;
    procedure CalcDefaultUsage;
    procedure StreamValues(const Streamer: TWStreamer); override;
    property DIndex : integer read fDIndex;
    property DName : String read GetdName write SetDName;
    property DSrcPic  : String read fDSrcPic write SetDSrcPic;
    property DSrcSize : integer read fDSrcSize write SetDSrcSize;
    property DSrcType : TWCSFlatFileSrcType read fDSrcType write SetDSrcType;
    property DSrcFilter  : String read fDSrcFilter write SetDSrcFilter;
    function UniqueDefName: String;
    procedure PicFromSize;

    // TOTAL Length in DECIMAL digits (not including dp/sign)
    property DSrcPrec : integer read fDSrcPrec write SetDSrcPrec;
    // Number of DP accuracy stored (<= Prec)
    property DSrcDP : integer read fDSrcDP write SetDSrcDP;
    property DSrcSignStorage : TWCSFlatFileNumFlag read fDSrcSignStorage write SetDSrcSignStorage;
    property DSrcDPStorage : TWCSFlatFileNumFlag read fDSrcDPStorage write SetDSrcDPStorage;
    // Translation Mask from Source to Dest e.g.
    // 9(6) time might be "hhmmss"
    // S9(9) COMP date might be be "yyyymmdd"
    property DSrcDstMask : string read fDSrcDstMask write SetDSrcDstMask;
    // Destination does not have an absolute size (dictated by internal storage)
    property DUsgType : TWCSFlatFileUsgType read fDUsgType write SetDUsgType;
    // String size or (for BCD) TOTAL Length in DECIMAL digits (not including dp/sign)
    property DDstID : Longint read fDDstID write SetDDstID;
  end;

  TFFFamily =
   (ffdf_Unknown    // used as a flag that the type has not been set
   ,ffdf_MFCobol    // All Microfocus Cobol
   ,ffdf_Generic    // Unknown file type (Always Unknown sub type)
   ,ffdf_Delimited  // All delimited text (Always Unknown sub type)
   ,ffdf_Btrieve    // All btrieve formats (subtypes to be defined)
   ,ffdf_XML        // All XML formats (subtypes to be defined)

    // --- add above here ------------------
    // new for streaming unsupported types - all resolve to this.
    // It is never stored so keep at the bottom.
   ,ffdf_Unsupported    // new for streaming unsupported types - all resolve to this
   );

  TFFSubfileType =
    (ffct_Unknown  // also generic for those with only 1 per family
    ,ffct_IndexedSequential
    ,ffct_RecordSequentialFixed
    ,ffct_RecordSequentialVariable
    ,ffct_RelativeFixed
    ,ffct_RelativeVariable
    ,ffct_LineSequential
    ,ffct_PrinterSequential

    // --- add above here ------------------
    // new for streaming unsupported types - all resolve to this.
    // It is never stored so keep at the bottom.
    ,ffct_Unsupported
    );
  TFFSubfileTypes = set of TFFSubfileType;

// Storage Mode (calculated from family and SubFileType)
  TWCSFlatRecType =
    (wffl_Unsupported      // For unsupported types - all resolve to this
    ,wffl_Fixed        // All records the same
    ,wffl_Variable     // Part of RecHead dictates the size
    ,wffl_Delimited    // Continues until specified Char
    ,wffl_Parsed       // OnParse returns field offsets and Lengths PER record
    );
  TWCSFlatQuoteType =
    (wffqOpenName
    ,wffqCloseName
    ,wffqOpenData
    ,wffqCloseData
    );
  TWCSFlatQuoteStore = array[TWCSFlatQuoteType] of Char;

  TFieldDefinitionList = class(TWcsDataObject)
  private
    // stored in stream
    fHeader: Longint; // also used to store lines of delimited header
    fFooter: Longint; // also used to store lines of delimited footer
    fRecLen: Longint;
    fRecHead: Longint;
    fRecFoot: Longint; // also used to store Quotes!
    fFillerItem: TFieldDefinition;
    fRecNoDstID: Longint;
    fFltNoDstID: Longint;
    fSubFileType: TFFSubFileType;
    fFamily: TFFFamily;  // this may be inherited from a database level param

    fRecDelim: String;
    fFldDelim: String;

    fHeaderLines: Longint;
    fFooterLines: Longint;
    fRecQuotes: Longint; // used to store Quotes as an integer for easy streaming!

    // calculated, run-time only
    fList: TList;

    procedure SetFooter(const Value: Longint);
    procedure SetHeader(const Value: Longint);
    procedure SetRecLen(const Value: Longint);
    procedure SetRecHead(const Value: Longint);
    procedure SetRecFoot(const Value: Longint);
    function GetItems(const Index: integer): TFieldDefinition;
    procedure ItemModified(Sender: TObject);
    function GetCount: integer;
    procedure SetSubFileType(const Value: TFFSubfileType);
    function GetMainFilterCount: Byte;
    procedure SetFltNoDstID(const Value: Longint);
    procedure SetRecNoDstID(const Value: Longint);
    function GetUsableLength: integer;
    procedure SetUsableLength(const Value: integer);
    procedure SetFamily(const Value: TFFFamily);
    procedure SetFldDelim(const Value: String);
    procedure SetRecDelim(const Value: String);
    function GetRecordType: TWCSFlatRecType;
    function GetDelimQuotes: TWCSFlatQuoteStore;
    procedure SetDelimQuotes(const Value: TWCSFlatQuoteStore);
    procedure SetFooterLines(const Value: Longint);
    procedure SetHeaderLines(const Value: Longint);
  protected
    procedure StreamValuesOLD(const Stream: TStream); override;
    procedure StreamValues(const Streamer: TWStreamer); override;
    procedure StreamResolve; override;
    procedure SetUnsaved(const Value: Boolean); override;
    procedure CreateSubClasses; override;
    procedure FreeSubClasses; override;
    function CreateSubData: TWCSDataObject; override;
    function GetDataObjName: String; override;
  public
    constructor Create; reintroduce;
    destructor Destroy; override;
    procedure Clear(const AndReinitialize: Boolean); override;
    procedure Init; override;
    function Add: integer;
    function AddDef(const AName: String): TFieldDefinition;
    function Indexof(const AName: String): integer;
    function DefByName(const AName: String): TFieldDefinition;
    function IndexofID(const AnID: Integer): integer;
    procedure RecalcFiller;
    procedure SetQuotes(const qOpenName,qCloseName,qOpenData,qCloseData : Char);

    procedure Insert(const Index: integer);
    procedure Delete(const Index: integer);
    procedure Exchange(const Index1, Index2: integer);
    function HasFields: Boolean;
    function FullFilterStr: String;
    property Count: integer read GetCount;
    function IDCount: integer;
    function UsedLength: integer;
    function RecCountEstimate(const FileSize: Int64): Extended;
    property UsableLength: integer read GetUsableLength write SetUsableLength;
    procedure GetFieldNoandPosn(const RecPosn: integer; var FieldNo, FieldPosn, FieldStart : integer);
    function FieldStartPosn(const FieldNo: integer): integer;
    property SubFileType: TFFSubFileType read fSubFileType write SetSubFileType;
    property MainFilterCount: Byte read GetMainFilterCount;
    property Items[const Index: integer]: TFieldDefinition read GetItems; default;
    property FillerItem: TFieldDefinition read fFillerItem;
    property Header: Longint read fHeader write SetHeader;
    property Footer: Longint read fFooter write SetFooter;
    property RecLen: Longint read fRecLen write SetRecLen;
    property RecHead: Longint read fRecHead write SetRecHead;
    property RecFoot: Longint read fRecFoot write SetRecFoot;
    property RecNoDstID: Longint read fRecNoDstID write SetRecNoDstID;
    property FltNoDstID: Longint read fFltNoDstID write SetFltNoDstID;
    property Family  : TFFFamily read fFamily   write SetFamily  ;
    property RecDelim: String    read fRecDelim write SetRecDelim;
    property FldDelim: String    read fFldDelim write SetFldDelim;
    property RecordType: TWCSFlatRecType read GetRecordType;
    property DelimQuotes: TWCSFlatQuoteStore read GetDelimQuotes write SetDelimQuotes;
    property HeaderLines: Longint read fHeaderLines write SetHeaderLines;
    property FooterLines: Longint read fFooterLines write SetFooterLines;

  end;


const
  cFlatFillerName = '<*Interal Filler*>';
  cFlatFileDecimals = [wffs_NumPk02, wffs_NumPk10, wffs_NumUnPk];
  cFlatFileFloats   = [wffs_Single, wffs_Double];
  cFlatFileNumerics = cFlatFileDecimals + cFlatFileFloats;

  sWCSFlatFileSrcType : array[TWCSFlatFileSrcType] of String =
     ('AlphaNum'
     ,'BinaryNum'
     ,'PackedNum'
     ,'Numeric'
     ,'FloatSingle'
     ,'FloatDouble'
     ,'Binary'
     );
  sWCSFlatFileUsgType : array[TWCSFlatFileUsgType] of String =
     ('AlphaNum'
     ,'LongInt'
     ,'LargeInt'
     ,'Float'
     ,'BCD'
     ,'DateTime'
     ,'Binary'
     );

  DigitUnSignedBytes : array[0..39] of byte =
    ( 0, 1, 1, 2, 2, 3, 3, 4, 4, 4,
      5, 5, 6, 6, 6, 7, 7, 8, 8, 9,
      9, 9,10,10,11,11,11,12,12,13,
     13,13,14,14,15,15,16,16,16,17);
  DigitSignedBytes : array[0..39] of byte =
    ( 0, 1, 1, 2, 2, 3, 3, 3, 4, 4,
      5, 5, 5, 6, 6, 7, 7, 8, 8, 8,
      9, 9,10,10,10,11,11,12,12,13,
     13,13,14,14,15,15,15,16,16,17);

  SFFFamily : array[TFFFamily] of string =
   ('Unassigned'
   ,'MicroFocus COBOL'
   ,'Generic fixed Length'
   ,'Delimited Text'
   ,'Btrieve'
   ,'XML'
   ,'Unsupported'
   );

  SFFSubFileType : array[TFFSubfileType] of string =
    ('Unknown/Generic'
    ,'Indexed Sequential'
    ,'Record Sequential Fixed'
    ,'Record Sequential Variable'
    ,'Relative Fixed'
    ,'Relative Variable'
    ,'Line Sequential'
    ,'Printer Sequential'
    ,'Unsupported'
    );

  cFFFamilySubTypes : array[TFFFamily] of TFFSubfileTypes =
   ([]//ffdf_Unknown
   ,[ffct_Unknown, ffct_IndexedSequential..ffct_PrinterSequential]//ffdf_MFCobol
   ,[ffct_Unknown] //ffdf_Generic
   ,[ffct_Unknown] //ffdf_Delimited
   ,[]//ffdf_Btrieve
   ,[]//ffdf_XML
   // --- add above here ------------------
   ,[ffct_UnSupported] //ffdf_Unsupported
   );

var
// This is how many Bits a DECIMAL number of n digits takes up.
// Add 7 (or 8 for Signed) and Divide by 8 to get Byte requirement for Packed COMP.
// Unit initialization Pre-fills this.
  DigitBits : array[0..39] of byte;

function StrToDateMaskRec(const Mask: String): TDateMaskRec;
function ExtractMaskedDate(const S: String; const Mask: TDateMaskRec; var DateRes: TDateTime): Boolean;
function CalcFlatFileRecordType(const FFFamily: TFFFamily; const FFSubFileType: TFFSubFileType): TWCSFlatRecType;
function FFFamilyFromStrObjs(const SL: TStrings; const Index: integer): TFFFamily;
function FFSubTypeFromStrObjs(const SL: TStrings; const Index: integer): TFFSubFileType;
function StrObjsFromFFFamily(const SL: TStrings;
  const IncludeUnknown: Boolean = False;
  const IncludeUnsupported: Boolean = False): integer;
function StrObjsFromFFSubType(const Family: TFFFamily; const SL: TStrings): integer;

implementation
uses
  WcsVarInt;
const
  DefRecQuotes: Longint = $27272222; // doubleQx2 SingleQx2
  DefRecDelim : String = #13#10;
  DefFldDelim : String = #9;

procedure FillDigitBits;
var
  Digit : integer;
  DigiVal: TVarInt;
  Digi10: TVarInt;
begin
  DigitBits[0] := 1;
  DigitBits[1] := 1;
  DigitBits[2] := 1;
  DigiVal:= TVarInt.Create(256);
  Digi10:= TVarInt.Create(8);
  try
    Digi10.viSetValue(10);
    DigiVal.viSetValue(100);
    for Digit := 3 to 39 do
    begin
      DigiVal.viMultiply(Digi10);
      DigitBits[Digit] := DigiVal.viHighBit+1;
    end;
  finally
    DigiVal.Free;
    Digi10.Free;
  end;
end;

function StrToDateMaskRec(const Mask: String): TDateMaskRec;
const
  DatePartChar : array[TDatePartType] of Char = ('Y','M','D','H','N','S');
var
  SS, SE, MP : integer;
  DP : TDatePartType;
begin
  Result.Valid := Mask > '';
  if not Result.Valid then Exit;
  with Result do
  begin
    Seps := False;
    for DP := low(TDatePartType) to High(TDatePartType) do
      Posns[DP] := 0;
    for MP := 1 to Length(Mask) do
      if Mask[MP] in ['/',' ',':'] then
        Seps := True
      else
        for DP := low(TDatePartType) to High(TDatePartType) do
          if Mask[MP] = DatePartChar[DP] then
            if Posns[DP] = 0 then
              Posns[DP] := MP
            else
              if (DP = dptYear) then
                YearLen := MP-Posns[DP]+1;
  end;                
end;

function ExtractMaskedDate(const S: String; const Mask: TDateMaskRec; var DateRes: TDateTime): Boolean;
var
  SS, SE : integer;

  function GetPart(const Posn, Len, Def: integer): Word;
  var
    SP, TP : integer;
  begin
    if (Posn > 0) and (SS+Posn <= SE) then
    begin
      Result := 0;
      SP := SS+Posn;
      for TP := 1 to Len do
        if (SP <= SE) and (S[SP] in ['0'..'9']) then
        begin
          Result := Result * 10 + Ord(S[SP])-48;
          inc(SP);
        end else
          Break;
    end else
      Result := Def;
  end;

  procedure AddTimePart(const TimePosn: integer; const TimeFactor: Double);
  begin
    if TimePosn > 0 then
      DateRes := DateRes + GetPart(TimePosn, 2, 0) / TimeFactor;
  end;
var
  Year: Word;
begin
  DateRes := 0;
  Result := False;
  if not Mask.Valid then Exit;
  SE := Length(S);
  SS := 1;
  while (SS <= SE) and (S[SS] in [#0..#32]) do inc(SS);
  while (SE >= SS) and (S[SE] in [#0..#32]) do dec(SE);
  if SE <= SS then Exit;
  Dec(SS); // make an Offset to add
  with Mask do
  begin
    // analyse Date.....
    if Posns[dptYear] > 0 then
    begin
      Year := GetPart(Posns[dptYear], YearLen, 0);
      if YearLen = 2 then
        if Year > 50 then inc(Year, 1900) else inc(Year, 2000);
      if not TryEncodeDate(Year,
                           GetPart(Posns[dptMon], 2, 1),
                           GetPart(Posns[dptDay], 2, 1),
                           DateREs) then
        Exit;
    end;
    // analyse Time.....
    if Posns[dptHour] > 0 then
    begin
      AddTimePart(Posns[dptHour], HoursPerDay);
      AddTimePart(Posns[dptMin], MinsPerDay);
      AddTimePart(Posns[dptSec], SecsPerDay);
    end;
  end;
  Result := True;
end;

function CalcFlatFileRecordType(const FFFamily: TFFFamily; const FFSubFileType: TFFSubFileType): TWCSFlatRecType;
begin
  if (FFFamily = ffdf_Unsupported) or (FFSubFileType = ffct_Unsupported) then
  begin
    Result := wffl_Unsupported;
    Exit;
  end;
  case FFFamily of
    ffdf_Unknown    : Result := wffl_Unsupported;
    ffdf_MFCobol    :
      begin
        if FFSubfileType in [ffct_RecordSequentialVariable,
                             ffct_RelativeVariable] then
          Result := wffl_Variable
        else
          Result := wffl_Fixed;
      end;
    ffdf_Generic    : Result := wffl_Fixed;
    ffdf_Delimited  : Result := wffl_Delimited;
    ffdf_Btrieve    :
      begin
      //!!!!! Need subtypes.....
        Result := wffl_Fixed;
      end;
    ffdf_XML        :
      begin
        Result := wffl_Parsed;
      end;
  end;
end;

function StrObjsFromFFFamily(const SL: TStrings;
  const IncludeUnknown: Boolean = False;
  const IncludeUnsupported: Boolean = False): integer;
var
  i : TFFFamily;
begin
  SL.Clear;
  if IncludeUnknown then SL.AddObject(SFFFamily[ffdf_Unknown], Pointer(ffdf_Unknown));
  for i := Succ(Low(TFFFamily)) to Pred(High(TFFFamily)) do
    SL.AddObject(SFFFamily[i], Pointer(i));
  if IncludeUnsupported then SL.AddObject(SFFFamily[ffdf_UnSupported], Pointer(ffdf_UnSupported));
  if SL.Count > 0 then
    Result := 0
  else
    Result := -1;
end;

function StrObjsFromFFSubType(const Family: TFFFamily;
  const SL: TStrings): integer;
var
  FFSLoop : TFFSubfileType;
begin
  SL.Clear;
  for FFSLoop := Low(TFFSubfileType) to High(TFFSubfileType) do
    if FFSLoop in cFFFamilySubTypes[Family] then
      SL.AddObject(SFFSubFileType[FFSLoop], Pointer(FFSLoop));
  if SL.Count > 0 then
    Result := 0
  else
    Result := -1;
end;

function FFFamilyFromStrObjs(const SL: TStrings; const Index: integer): TFFFamily;
begin
  if (Index >= 0) and (Index <= SL.Count) then
    Result := TFFFamily(SL.Objects[Index])
  else
    Result := ffdf_Unknown;
end;

function FFSubTypeFromStrObjs(const SL: TStrings; const Index: integer): TFFSubFileType;
begin
  if (Index >= 0) and (Index <= SL.Count) then
    Result := TFFSubFileType(SL.Objects[Index])
  else
    Result := ffct_Unknown;
end;

{ TFieldDefinition }

procedure TFieldDefinition.Init;
begin
  fDName := '';
  fDSrcSize:= 0;
  fDUsgType:= wffu_Chars;
  fDSrcType:= wffs_Chars;
end;

constructor TFieldDefinition.Create(const AList: TFieldDefinitionList);
begin
  if AList <> Nil then
    inherited Create(AList.WriteStreamVersion)
  else
    inherited Create(cFFListVers);
  StringLimit := 255;
  fdList := AList;
  if AList <> Nil then
    OnAfterPost := fdList.ItemModified;
end;

function TFieldDefinition.GetdName: String;
begin
  if fdName > '' then
    Result := fdName
  else
    Result := Format('[%3.3d] (Un-named)', [fdIndex]);
end;

procedure TFieldDefinition.SetDName(const Value: String);
begin
  if (Value <> fDName) then
  begin
    if Value = GetDName then Exit;
    SetStrValue(Value, fDName);
  end;
end;

procedure TFieldDefinition.SetDSrcSize(const Value: integer);
begin
  SetIntValue(Value, fDSrcSize);
end;

procedure TFieldDefinition.SetDSrcType(const Value: TWCSFlatFileSrcType);
begin
  SetEnumValue(Value, fDSrcType);
end;

procedure TFieldDefinition.SetDUsgType(const Value: TWCSFlatFileUsgType);
begin
  SetEnumValue(Value, fDUsgType);
end;

procedure TFieldDefinition.SetDDstID(const Value: Longint);
begin
  SetIntValue(Value, fDDstID);
end;

procedure TFieldDefinition.SetDSrcDP(const Value: integer);
begin
  SetIntValue(Value, fDSrcDP);
end;

procedure TFieldDefinition.SetDSrcDPStorage(const Value: TWCSFlatFileNumFlag);
begin
  SetEnumValue(Value, fDSrcDPStorage);
end;

procedure TFieldDefinition.SetDSrcDstMask(const Value: string);
begin
  SetStrValue(Value, fDSrcDstMask);
end;

procedure TFieldDefinition.SetDSrcPrec(const Value: integer);
begin
  SetIntValue(Value, fDSrcPrec);
end;

procedure TFieldDefinition.SetDSrcSignStorage(const Value: TWCSFlatFileNumFlag);
begin
  SetEnumValue(Value, fDSrcSignStorage);
end;

procedure TFieldDefinition.SetDSrcPic(const Value: String);
begin
  SetStrValue(Value, fDSrcPic);
end;

procedure TFieldDefinition.SetDSrcFilter(const Value: String);
begin
  SetStrValue(Value,fDSrcFilter);
end;

procedure TFieldDefinition.StreamValuesOLD(const Stream: TStream);
begin
  StreamStrValueOLD(Stream, fdName);
  Stream.Read(fdSrcSize, 4);
  Stream.Read(fdSrcType, 1);
  Stream.Read(fdUsgType, 1);
  if (ReadStreamVersion >= 2) then
  begin
    Stream.Read(fdDstID, 4);
    if (ReadStreamVersion >= 3) then
    begin
      StreamStrValueOLD(Stream, fDSrcPic    );
      StreamStrValueOLD(Stream, fDSrcDstMask);
      if fdSrcType in cFlatFileNumerics then
      begin
        Stream.Read(fDSrcPrec , 4);
        Stream.Read(fDSrcDP   , 4);
        if fdSrcType in cFlatFileDecimals then
        begin
          Stream.Read(fDSrcSignStorage, 1);
          Stream.Read(fDSrcDPStorage  , 1);
        end;
      end else
        fDSrcPrec := fDSrcSize;
    end;
    if (ReadStreamVersion >= 6) then StreamStrValueOLD(Stream, fDSrcFilter);
  end;
end;

procedure TFieldDefinition.StreamValues(const Streamer: TWStreamer);
var
  Dummy: String; i: integer;
begin
  with Streamer do
  begin
    StreamString(fdName);
    StreamLongint(fdSrcSize);
    StreamEnum(fdSrcType);
    StreamEnum(fdUsgType);
    StreamLongint(fdDstID);
    StreamString(fDSrcPic     );
    StreamString(fDSrcDstMask );
    if fdSrcType in cFlatFileNumerics then
    begin
      StreamLongint(fDSrcPrec );
      StreamLongint(fDSrcDP   );
      if fdSrcType in cFlatFileDecimals then
      begin
        StreamEnum(fDSrcSignStorage);
        StreamEnum(fDSrcDPStorage  );
      end;
    end else
      if not Writing then fDSrcPrec := fDSrcSize;
    StreamString(fDSrcFilter );

   // - Post V8 extensions HERE
{  This for testing bi-directional streaming only!!!
    if (CurrStreamVersion[Writing] >=  xxx) then
    begin
      i := 8;
      StreamEnum(i);
      StreamLongint(i);
      StreamString(Dummy);
      StreamString(Dummy);
      StreamString(Dummy);
    end;
}
  end;
end;

procedure TFieldDefinition.CalcDefaultUsage;
begin
  case fDSrcType of
    wffs_Chars     :
      begin
        fDUsgType := wffu_Chars;
      end;
    wffs_NumPk02   :
      begin
        if fDSrcDP = 0 then
        begin
          if fDSrcSize <= 4 then
            fDUsgType := wffu_LongInt
          else if fDSrcSize <= 8 then
            fDUsgType := wffu_LargeInt
          else fDUsgType := wffu_BCD;
        end else begin
          if fDSrcPrec <= 16 then
            fDUsgType := wffu_Float
          else
            fDUsgType := wffu_BCD;
        end;
      end;
    wffs_NumPk10, wffs_NumUnPk :
      begin
        if fDSrcDP = 0 then
        begin
          if fDSrcPrec <= 9 then
            fDUsgType := wffu_LongInt
          else if fDSrcPrec <= 18 then
            fDUsgType := wffu_LargeInt
          else fDUsgType := wffu_BCD;
        end else begin
          if fDSrcPrec <= 16 then
            fDUsgType := wffu_Float
          else
            fDUsgType := wffu_BCD;
        end;
      end;
    wffs_Single    :
      begin
        fDUsgType := wffu_Float
      end;
    wffs_Double    :
      begin
        fDUsgType := wffu_Float
      end;
    wffs_Binary    :
      begin
        fDUsgType := wffu_Binary;
      end;
  end;
end;

function TFieldDefinition.GetDataObjName: String;
begin
  Result := fDName;
end;

function TFieldDefinition.UniqueDefName: String;
var
  i, j: integer;
begin
  j := 0;
  for i := 0 to fDIndex-1 do
    if CompareText(fDList[i].fDName, fDName) = 0 then
      inc(j);
  if fDName > '' then
    Result := fDName
  else
    Result := 'UnNamed';
  if j > 0 then
    Result := Result + '_' + IntToStr(j);
end;

procedure TFieldDefinition.PicFromSize;
begin
  fDSrcPic := Format('X(%d)',[fDSrcSize]);
end;

{ TFieldDefinitionList }

procedure TFieldDefinitionList.CreateSubClasses;
begin
  fList := TList.Create;
  fFillerItem:= TFieldDefinition.Create(Self);
  fFillerItem.fDName := cFlatFillerName;
  fFillerItem.fDDstID := -2;
  fFillerItem.fDIndex := -1;
end;

constructor TFieldDefinitionList.Create;
begin
  inherited Create(cFFListVers);
end;

procedure TFieldDefinitionList.FreeSubClasses;
begin
  fList.Free;
  fFillerItem.Free;
end;

destructor TFieldDefinitionList.Destroy;
begin
  inherited;
end;

function TFieldDefinitionList.CreateSubData: TWCSDataObject;
var
  NewItem: TFieldDefinition;
begin
  NewItem:= TFieldDefinition.Create(Self);
  LinkChild(NewItem);
  NewItem.fdIndex := fList.Count;
  inc(fFillerItem.fDIndex);
  Result := NewItem;
end;

procedure TFieldDefinitionList.Insert(const Index: integer);
var
  NewItem: TFieldDefinition;
  i : integer;
begin
  NewItem:= TFieldDefinition(CreateSubData);
  NewItem.fdIndex := Index;
  fList.Insert(Index, NewItem);
  for i := Index+1 to fList.Count-1 do
    TFieldDefinition(fList[i]).fdIndex := i;
end;

procedure TFieldDefinitionList.Delete(const Index: integer);
var
  i : integer;
begin
  if Index = fList.Count then
    raise exception.Create('Filler item cannot be deleted');
  TFieldDefinition(fList[Index]).Free;
  fList.Delete(Index);
  for i := Index to fList.Count-1 do
    TFieldDefinition(fList[i]).fdIndex := i;
end;

function TFieldDefinitionList.Add: integer;
begin
  Result := fList.Add(CreateSubData);
end;

function TFieldDefinitionList.AddDef(const AName: String): TFieldDefinition;
var
  i : integer;
begin
  i := Add;
  Result := TFieldDefinition(fList[i]);
  Result.fDName := AName;
end;

procedure TFieldDefinitionList.Init;
var
  FillerItem: TFieldDefinition;
  i : integer;
begin
  fHeader := 0;
  fRecLen := 0;
  fFooter := 0;
  fRecHead := 0;
  fFillerItem.fDSrcSize := 0;
  fFillerItem.fDIndex := 0;
  fRecNoDstId := 0;
  fFltNoDstId := 0;
  fRecQuotes := DefRecQuotes;
  fRecDelim := DefRecDelim;
  fFldDelim := DefFldDelim;
end;

procedure TFieldDefinitionList.Clear(const AndReinitialize: Boolean);
var
  i : integer;
begin
  for i := fList.Count - 1 downto 0 do
  begin
    TFieldDefinition(fList[i]).Free;
    fList.Delete(i);
  end;
  inherited;
end;

function TFieldDefinitionList.HasFields: Boolean;
begin
  Result := fList.Count > 0;
end;

procedure TFieldDefinitionList.Exchange(const Index1, Index2: integer);
begin
  fList.Exchange(Index1, Index2);
end;

function TFieldDefinitionList.GetCount: integer;
begin
  Result := fList.Count + 1;
end;

function TFieldDefinitionList.GetUsableLength: integer;
begin
  Result := fRecLen - fRecHead - fRecFoot;
end;

procedure TFieldDefinitionList.SetUsableLength(const Value: integer);
begin
  RecLen := Value + fRecHead + fRecFoot;
end;

function TFieldDefinitionList.UsedLength: integer;
var
  i : integer;
begin
  Result := 0;
  for i := 0 to fList.Count - 1 do
    inc(Result, TFieldDefinition(fList[i]).fDSrcSize);
end;

procedure TFieldDefinitionList.RecalcFiller;
begin
  fFillerItem.fDSrcSize := fRecLen - UsedLength - fRecHead - fRecFoot;
  fFillerItem.PicFromSize;
end;

function TFieldDefinitionList.GetItems(const Index: integer): TFieldDefinition;
begin
  if index = fList.Count then
    Result := fFillerItem
  else
    Result := TFieldDefinition(fList[Index]);
end;

procedure TFieldDefinitionList.StreamValuesOLD(const Stream: TStream);
var
  Dummy: String;
begin
  Stream.Read(fRecLen , 4);
  Stream.Read(fHeader , 4);
  Stream.Read(fFooter , 4);
  Stream.Read(fRecHead, 4);
  if (ReadStreamVersion >= 4) then
    Stream.Read(fRecFoot, 4)
  else
    fRecFoot := 0;
  if (ReadStreamVersion >= 5) then
  begin
    Stream.Read(fSubFileType, 1);
    Stream.Seek(3, soCurrent);
    if (ReadStreamVersion < 6) then
      StreamStrValueOld(Stream, Dummy);
    if (ReadStreamVersion >= 7) then
    begin
      Stream.Read(fRecNoDstId, 4);
      Stream.Read(fFltNoDstId, 4);
    end;
  end else
    fSubFileType := ffct_Unknown;
  StreamSubValueListOld(Stream, fList, 1024);
  RecalcFiller;
end;

procedure TFieldDefinitionList.StreamValues(const Streamer: TWStreamer);
var
  i : integer;
  Dummy: String;
begin
  with Streamer do
  begin
    if (ReadStreamVersion = 0) and (not Writing) then CorrectReadStreamVersion(8);
    StreamLongInt(fRecLen );
    StreamLongInt(fHeader );
    StreamLongInt(fFooter );
    StreamLongInt(fRecHead);
    StreamLongInt(fRecFoot);
    StreamEnumProtect(fSubFileType, ord(High(TFFSubfileType)),
                      ord(ffct_Unsupported), WStrmLongint1);
    StreamLongInt(fRecNoDstId );
    StreamLongInt(fFltNoDstId );
    StreamObjectList(fList, StreamSubValues, 1024);

  // - Post V8 extensions HERE
    if (CurrStreamVersion[Writing] >=  9) then
    begin
      StreamEnumProtect(fFamily, ord(High(TFFFamily)), ord(ffdf_Unsupported));
      StreamEnum(fFamily);
      StreamString(fRecDelim);
      StreamString(fFldDelim);
    end else begin
      fRecDelim := DefRecDelim;
      fFldDelim := DefFldDelim;
    end;

    if (CurrStreamVersion[Writing] >= 10) then
    begin
      StreamLongInt(fRecQuotes);
      StreamLongInt(fHeaderLines);
      StreamLongInt(fFooterLines);
    end else
      fRecQuotes := DefRecQuotes;

{  This for testing bi-directional streaming only!!!
    if (CurrStreamVersion[Writing] >=  xx) then
    begin
      i := 8;
      StreamEnum(i);
      StreamLongint(i);
      StreamString(Dummy);
      StreamString(Dummy);
      StreamString(Dummy);
    end;
}
  end;
end;

procedure TFieldDefinitionList.StreamResolve;
begin
  RecalcFiller;
end;

procedure TFieldDefinitionList.ItemModified(Sender: TObject);
begin
//
end;

procedure TFieldDefinitionList.GetFieldNoandPosn(const RecPosn: integer;
  var FieldNo, FieldPosn, FieldStart: integer);
begin
  FieldStart := fRecHead;
  FieldNo := 0;
  while FieldNo < fList.Count do
  with TFieldDefinition(fList[FieldNo]) do
  begin
    if FieldStart + fdSrcSize > RecPosn then
      Break;
    inc(FieldStart, fdSrcSize);
    inc(FieldNo);
  end;
  FieldPosn := RecPosn - FieldStart;
  dec(FieldStart, fRecHead);
end;

function TFieldDefinitionList.FieldStartPosn(const FieldNo: integer): integer;
var
  i : integer;
begin
  Result := fRecHead;
  for i := 0 to FieldNo-1 do
    inc(Result, TFieldDefinition(fList[i]).fDSrcSize);
end;

procedure TFieldDefinitionList.SetUnsaved(const Value: Boolean);
var
  i : integer;
begin
  if Value <> Unsaved then
  begin
    inherited;
    if not Value then
    begin
      for i := 0 to fList.Count - 1 do
        TFieldDefinition(fList[i]).UnSaved := False;
      fFillerItem.UnSaved := False;
    end;
  end;
end;

procedure TFieldDefinitionList.SetFooter(const Value: Longint);
begin
  SetIntValue(Value, fFooter);
end;

procedure TFieldDefinitionList.SetHeader(const Value: Longint);
begin
  SetIntValue(Value, fHeader);
end;

procedure TFieldDefinitionList.SetRecLen(const Value: Longint);
begin
  SetIntValue(Value, fRecLen);
  RecalcFiller;
end;

procedure TFieldDefinitionList.SetRecHead(const Value: Longint);
begin
  SetIntValue(Value, fRecHead);
  RecalcFiller;
end;

procedure TFieldDefinitionList.SetRecFoot(const Value: Longint);
begin
  SetIntValue(Value, fRecFoot);
  RecalcFiller;
end;

procedure TFieldDefinitionList.SetSubFileType(const Value: TFFSubfileType);
begin
  SetENumValue(Value, fSubFileType);
end;

procedure TFieldDefinitionList.SetFltNoDstID(const Value: Longint);
begin
  SetIntValue(Value, fFltNoDstID);
end;

procedure TFieldDefinitionList.SetRecNoDstID(const Value: Longint);
begin
  SetIntValue(Value, fRecNoDstID);
end;

procedure TFieldDefinitionList.SetFamily(const Value: TFFFamily);
begin
  if SetENumValue(Value, fFamily) then
  begin
  end;
end;

procedure TFieldDefinitionList.SetFldDelim(const Value: String);
begin
  SetStrValue(Value, fFldDelim);
end;

procedure TFieldDefinitionList.SetRecDelim(const Value: String);
begin
  SetStrValue(Value, fRecDelim);
end;

function TFieldDefinitionList.GetDataObjName: String;
begin
  Result := 'FieldDefList';
end;

function TFieldDefinitionList.GetMainFilterCount: Byte;
var
  i : integer;
begin
  Result := 0;
  for i := 0 to fList.Count - 1 do
    if Items[i].fDSrcFilter > '' then
      inc(Result);
end;

function TFieldDefinitionList.FullFilterStr: String;
var
  i, j : integer;
begin
  Result := '';
  j := 0;
  for i := 0 to fList.Count-1 do
  with Items[i] do
  if (fDSrcFilter > '') then
  begin
    case j of
      0 : Result := fDSrcFilter;
      1 : Result := '(' + Result + ') AND (' + fDSrcFilter + ')';
      else Result := Result + ' AND (' + fDSrcFilter + ')';
    end;
    inc(j);
  end;

end;

function TFieldDefinitionList.Indexof(const AName: String): integer;
begin
  for Result := 0 to fList.Count - 1 do
    if CompareText(TFieldDefinition(fList[Result]).DName, AName) = 0 then
      Exit;
  Result := -1;      
end;

function TFieldDefinitionList.IndexofID(const AnID: Longint): integer;
begin
  for Result := 0 to fList.Count - 1 do
    if TFieldDefinition(fList[Result]).DDstID = AnID then
      Exit;
  Result := -1;
end;

function TFieldDefinitionList.DefByName(const AName: String): TFieldDefinition;
var
  i : integer;
begin
  i := Indexof(AName);
  if i >= 0 then
    Result := TFieldDefinition(fList[i])
  else
    Result := Nil;
end;

function TFieldDefinitionList.IDCount: integer;
var
  i : integer;
begin
  Result := 0;
  for i := 0 to fList.Count - 1 do
    if TFieldDefinition(fList[i]).fDDstID > 0 then
      inc(Result);
end;

function TFieldDefinitionList.RecCountEstimate(const FileSize: Int64): Extended;
var
  AvailSize: Int64;
begin
  if RecLen > 0 then
  begin
    AvailSize := FileSize - fHeader - fFooter;
    Result := AvailSize / fRecLen;
  end else
    Result := -1;
end;

function TFieldDefinitionList.GetRecordType: TWCSFlatRecType;
begin
  Result := CalcFlatFileRecordType(fFamily, fSubFileType);
end;

function TFieldDefinitionList.GetDelimQuotes: TWCSFlatQuoteStore;
begin
  Result := TWCSFlatQuoteStore(fRecQuotes);
end;

procedure TFieldDefinitionList.SetDelimQuotes(const Value: TWCSFlatQuoteStore);
begin
  SetIntValue(Longint(Value), fRecQuotes);
end;

procedure TFieldDefinitionList.SetFooterLines(const Value: Longint);
begin
  SetIntValue(Value, fFooterLines);
end;

procedure TFieldDefinitionList.SetHeaderLines(const Value: Longint);
begin
  SetIntValue(Value, fHeaderLines);
end;

procedure TFieldDefinitionList.SetQuotes(const qOpenName, qCloseName,
  qOpenData, qCloseData: Char);
var
  QA: TWCSFlatQuoteStore;
begin
  QA[wffqOpenName ]  := qOpenName  ;
  QA[wffqCloseName]  := qCloseName ;
  QA[wffqOpenData ]  := qOpenData  ;
  QA[wffqCloseData]  := qCloseData ;
  SetDelimQuotes(QA);
end;

initialization
  FillDigitBits;
end.
