{
  Roland Potter - Woodstock Computer Solutions
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.

  A potentially enormous (but slow) variable length integer!
}

unit WcsVarInt;

interface

type
  TVarInt = class(TObject)
  private
    fHighBit: integer;
    fBitCount: integer;
    fByteCount: integer;
    fBytes: array of Byte;
    procedure viOverflow(const Op: String);
    procedure MoveHighBitDown(const MaxBitNo: integer);
    procedure viDiv(const AVarInt: TVarInt);
    procedure viMod(const AVarInt: TVarInt);
  public
    constructor Create(const ABitCount: integer);
    procedure viZero;
    function viIsZero: Boolean;
    procedure viAssign(const AVarInt: TVarInt);
    procedure viSetValue(AnInteger: Int64);
    property viBitCount: integer read fBitCount;
    property viByteCount: integer read fByteCount;
    property viHighBit: integer read fHighBit;
    function viBitSet(const BitNo: integer): Boolean;
    procedure viSetBit(const BitNo: integer);
    procedure viUnSetBit(const BitNo: integer);
    procedure viSHL;
    procedure viSHR;
    procedure viAdd(const AVarInt: TVarInt);
    procedure viMultiply(const AVarInt: TVarInt);
    procedure viSubtract(const AVarInt: TVarInt);
  end;

implementation
uses
  SysUtils;

{ TVarInt }

constructor TVarInt.Create(const ABitCount: integer);
begin
  fBitCount:= ABitCount;
  fByteCount:= (fBitCount + 7) div 8;
  SetLength(fBytes, fByteCount);
  viZero;
end;

procedure TVarInt.viOverflow(const Op: String);
begin
  raise exception.Create(IntToStr(fBitCount)+'-Bit Integer Overflow on '+Op);
end;

procedure TVarInt.viZero;
begin
  FillChar(fBytes[0], fByteCount, 0);
  fHighBit := -1;
end;

function TVarInt.viIsZero: Boolean;
begin
  Result := fHighBit < 0;
end;

procedure TVarInt.viAssign(const AVarInt: TVarInt);
begin
  if AVarInt <> Nil then
  begin
    if AVarInt.fBitCount > fBitCount then viOverflow('Assign');
    viZero;
    move(AVarInt.fBytes[0], fBytes[0], AVarInt.fByteCount);
    fHighBit := AVarInt.fHighBit;
  end else
    viZero;
end;

procedure TVarInt.MoveHighBitDown(const MaxBitNo: integer);
var
  ByteNo, BitNo : integer;
begin
  fHighBit := MaxBitNo;
  ByteNo := MaxBitNo div 8;
  BitNo  := MaxBitNo mod 8;
  while ByteNo >= 0 do
  begin
    while BitNo >= 0 do
    begin
      if fBytes[ByteNo] and (1 shl BitNo) > 0 then Exit;
      dec(fHighBit);
      dec(BitNo);
    end;
    BitNo := 7;
    dec(ByteNo);
  end;
end;

function TVarInt.viBitSet(const BitNo: integer): Boolean;
begin
  if (BitNo > fBitCount-1) or (BitNo < 0) then viOverflow('BitSet');
  if BitNo = fHighBit then
    Result := True
  else
    if BitNo > fHighBit then
      Result := False
    else
      Result := fBytes[BitNo div 8] and (1 shl (BitNo mod 8)) > 0;
end;

procedure TVarInt.viSetBit(const BitNo: integer);
begin
  if (BitNo > fBitCount-1) or (BitNo < 0) then viOverflow('SetBit');
  fBytes[BitNo div 8] := fBytes[BitNo div 8] or (1 shl (BitNo mod 8));
  if BitNo > fHighBit then fHighBit := BitNo;
end;

procedure TVarInt.viUnSetBit(const BitNo: integer);
begin
  if (BitNo > fBitCount-1) or (BitNo < 0) then viOverflow('UnSetBit');
  fBytes[BitNo div 8] := fBytes[BitNo div 8] and not (1 shl (BitNo mod 8));
  MoveHighBitDown(fHighBit-1);
end;

procedure TVarInt.viSetValue(AnInteger: Int64);
var
  MoveCount: integer;
begin
  if (fBitCount < 64) then
  begin
    if (AnInteger shr fBitCount <> 0) then viOverflow('SetValue');
    MoveCount := fByteCount;
  end else
    MoveCount := 8;
  viZero;
  Move(AnInteger, fBytes[0], MoveCount);
  if (fBitCount < 64) then
    MoveHighBitDown(fBitCount-1)
  else
    MoveHighBitDown(63);
end;

procedure TVarInt.viSHL;
var
  i, HighByte: integer;
begin
  if fHighBit >= fBitCount-1 then viOverflow('SHL');
  inc(fHighBit);
  HighByte := (fHighBit) div 8;
  fBytes[HighByte] := fBytes[HighByte] shl 1;
  for i := HighByte-1 downto 0 do
  begin
    if fBytes[i] and (1 shl 7) > 0 then
      fBytes[i+1] := fBytes[i+1] or 1;
    fBytes[i] := fBytes[i] shl 1;
  end;
end;

procedure TVarInt.viSHR;
var
  i : integer;
begin
  if fHighBit < 0 then Exit;
  fBytes[0] := fBytes[0] shr 1;
  for i := 1 to fHighBit div 8 do
  begin
    fBytes[i-1] := fBytes[i-1] or ((fBytes[i] and 1) shl 7);
    fBytes[i] := fBytes[i] shr 1;
  end;
  dec(fHighBit);
end;

procedure TVarInt.viSubtract(const AVarInt: TVarInt);
var
  HighByte,
  iByte, Carry, ByteRes : integer;
begin
  if AVarInt.fHighBit < 0 then Exit;
  if AVarInt.fHighBit > fBitCount-1 then viOverflow('SUB');
  HighByte := AVarInt.fHighBit div 8;
  iByte := 0;
  Carry := 0;
  while iByte <= HighByte do
  begin
    ByteRes := fBytes[iByte];
    ByteRes := ByteRes - AVarInt.fBytes[iByte] - Carry;
    if ByteRes >= 0 then
    begin
      fBytes[iByte] := ByteRes;
      Carry := 0;
    end else begin
      fBytes[iByte] := 256+ByteRes;
      Carry := 1;
    end;
    inc(iByte);
  end;
  while (Carry > 0) and (iByte < fByteCount) do
  begin
    ByteRes := fBytes[iByte]-Carry;
    if ByteRes >= 0 then
    begin
      fBytes[iByte] := ByteRes;
      Carry := 0;
    end else begin
      fBytes[iByte] := 256+ByteRes;
      Carry := 1;
    end;
    inc(iByte);
  end;
  MoveHighBitDown(fBitCount-1);
end;

procedure TVarInt.viAdd(const AVarInt: TVarInt);
var
  HighByte,
  iByte, Carry, ByteRes : integer;
begin
  if AVarInt.fHighBit < 0 then Exit;
  if AVarInt.fHighBit > fBitCount-1 then viOverflow('ADD');
  HighByte := AVarInt.fHighBit div 8;
  iByte := 0;
  Carry := 0;
  while iByte <= HighByte do
  begin
    ByteRes := fBytes[iByte];
    ByteRes := ByteRes + AVarInt.fBytes[iByte] + Carry;
    if ByteRes < 256 then
    begin
      fBytes[iByte] := ByteRes;
      Carry := 0;
    end else begin
      fBytes[iByte] := ByteRes-256;
      Carry := 1;
    end;
    inc(iByte);
  end;
  while (Carry > 0) and (iByte < fByteCount) do
  begin
    ByteRes := fBytes[iByte]+Carry;
    if ByteRes < 256 then
    begin
      fBytes[iByte] := ByteRes;
      Carry := 0;
    end else begin
      fBytes[iByte] := ByteRes-256;
      Carry := 1;
    end;
    inc(iByte);
  end;
  MoveHighBitDown(fBitCount-1);
end;

procedure TVarInt.viMultiply(const AVarInt: TVarInt);
var
  HighByte, HighBytev,
  iByte,iBytev, Carry, ByteRes : integer;
  TmpVarInt : TVarInt;
begin
  if AVarInt.fHighBit < 1 then
  begin
    if AVarInt.fHighBit < 0 then
      viZero; // x 0
    Exit; // x 1 => No Change
  end;
  if AVarInt.fHighBit+fHighBit > fBitCount-1 then viOverflow('MUL');
  TmpVarInt := TVarInt.Create(fBitCount);
  try
    HighByte := fHighBit div 8;
    HighBytev := AVarInt.fHighBit div 8;
    TmpVarInt.viAssign(Self);
    viZero;
    Carry := 0;
    iBytev := 0;
    while iBytev <= HighBytev do
    begin
      iByte := 0;
      while iByte <= HighByte do
      begin
        ByteRes := TmpVarInt.fBytes[iByte];
        ByteRes := (ByteRes * AVarInt.fBytes[iBytev]) + fBytes[iBytev + iByte] + Carry;
        if ByteRes < 256 then
        begin
          fBytes[iBytev + iByte] := ByteRes;
          Carry := 0;
        end else begin
          fBytes[iBytev + iByte] := (ByteRes) and $FF;
          Carry := (ByteRes) shr 8;
        end;
        inc(iByte);
      end;
      while (Carry > 0) and (iByte+iBytev < fByteCount) do
      begin
        ByteRes := TmpVarInt.fBytes[iByte];
        ByteRes := ByteRes + fBytes[iBytev + iByte] + Carry;
        if ByteRes < 256 then
        begin
          fBytes[iByte] := ByteRes;
          Carry := 0;
        end else begin
          fBytes[iByte] := (ByteRes) and $FF;
          Carry := (ByteRes) shr 8;
        end;
        inc(iByte);
      end;
      Inc(iBytev);
    end;
  finally
    TmpVarInt.Free;
  end;
  MoveHighBitDown(fBitCount-1);
end;

procedure TVarInt.viDiv(const AVarInt: TVarInt);
begin
  if AVarInt.fHighBit < 1 then
  begin
    if AVarInt.fHighBit < 0 then
      viOverflow('DIV');
    Exit; // Div 1 => no Change
  end;
  if AVarInt.fHighBit > fHighBit then
  begin
    viZero;
    Exit;
  end;
//
end;

procedure TVarInt.viMod(const AVarInt: TVarInt);
begin
  if AVarInt.fHighBit < 1 then
    viOverflow('MOD');
//
end;

end.
